/*
Copyright (C) 2024, PCB Artists OPC Pvt Ltd, all right reserved.

Author:     Pratik Panda
E-mail:     hello@pcbartists.com
Website:    pcbartists.com

The code referencing this license is open source software. Redistribution
and use of the code in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:
    - Redistribution of original and modified source code must retain the 
    above copyright notice, this condition and the following disclaimer.
    This code (or modifications) in source or binary forms may NOT be used
    in a commercial application without obtaining permission from the Author.
    - This software is provided by the copyright holder and contributors "AS IS"
    and any warranties related to this software are DISCLAIMED. The copyright 
    owner or contributors are NOT LIABLE for any damages caused by use of this 
    software.
*/

#include <stdio.h>
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "driver/gpio.h"
#include "driver/i2c.h"
#include "esp_log.h"
#include "sdkconfig.h"

#include "dbm_flex.h"

// I2C address of the decibel meter (0x48 is default)
#define     DBM_I2C_ADDR        0x48

// Register addresses
#define 	I2C_REG_VERSION		0x00

#define 	I2C_REG_ID3			0x01
#define 	I2C_REG_ID2			0x02
#define 	I2C_REG_ID1			0x03
#define 	I2C_REG_ID0			0x04

#define 	I2C_REG_RSVD		0x05

#define 	I2C_REG_CONTROL		0x06
#define 	I2C_REG_TAVG_HIGH	0x07
#define 	I2C_REG_TAVG_LOW	0x08
#define 	I2C_REG_RESET		0x09

#define 	I2C_REG_DBA_H		10
#define 	I2C_REG_DBA_L		11
#define 	I2C_REG_DBC_H		12
#define 	I2C_REG_DBC_L		13
#define 	I2C_REG_DBZ_H		14
#define 	I2C_REG_DBZ_L		15

#define 	I2C_REG_LAEQF_H		16
#define 	I2C_REG_LAEQF_L		17
#define 	I2C_REG_LCEQF_H		18
#define 	I2C_REG_LCEQF_L		19
#define 	I2C_REG_LZEQF_H		20
#define 	I2C_REG_LZEQF_L		21

#define 	I2C_REG_LAEQS_H		22
#define 	I2C_REG_LAEQS_L		23
#define 	I2C_REG_LCEQS_H		24
#define 	I2C_REG_LCEQS_L		25
#define 	I2C_REG_LZEQS_H		26
#define 	I2C_REG_LZEQS_L		27

#define 	I2C_REG_LAPK_H		28
#define 	I2C_REG_LAPK_L		29
#define 	I2C_REG_LCPK_H		30
#define 	I2C_REG_LCPK_L		31
#define 	I2C_REG_LZPK_H		32
#define 	I2C_REG_LZPK_L		33

#define 	I2C_REG_LAMAX_H		34
#define 	I2C_REG_LAMAX_L		35
#define 	I2C_REG_LCMAX_H		36
#define 	I2C_REG_LCMAX_L		37
#define 	I2C_REG_LZMAX_H		38
#define 	I2C_REG_LZMAX_L		39

#define 	I2C_REG_LAMIN_H		40
#define 	I2C_REG_LAMIN_L		41
#define 	I2C_REG_LCMIN_H		42
#define 	I2C_REG_LCMIN_L		43
#define 	I2C_REG_LZMIN_H		44
#define 	I2C_REG_LZMIN_L		45

#define 	I2C_REG_THR_MIN		46
#define 	I2C_REG_THR_MAX		47

#define 	I2C_REG_T_OVER2		48
#define 	I2C_REG_T_OVER1		49
#define 	I2C_REG_T_OVER0		50
#define 	I2C_REG_T_UNDER2	51
#define 	I2C_REG_T_UNDER1	52
#define 	I2C_REG_T_UNDER0	53

#define 	I2C_REG_HISTORY_SEC	100
#define 	I2C_REG_HISTORY_MIN	160
#define 	I2C_REG_HISTORY_HRS	220

static char *TAG = "dbm_flex.c";

/*
Function:   Write I2C registers on the decibel sensor
NOTE:       Can be replaced with your custom I2C write code
Args:       i2c,        i2c port
            reg,        register address to write to
            data,       byte array to write
            count,      number of bytes to write
Returns:    ESP_OK or ESP_FAIL
*/
esp_err_t i2cmaster_write_registers (i2c_port_t i2c, uint8_t reg, uint8_t *data, uint8_t count)
{
    // Create a new I2C command
    i2c_cmd_handle_t cmd = i2c_cmd_link_create();
    i2c_master_start(cmd);
    i2c_master_write_byte(cmd, (DBM_I2C_ADDR << 1) | I2C_MASTER_WRITE, true);
    i2c_master_write_byte(cmd, reg, true);
    i2c_master_write(cmd, data, count, true);
    // Stop the I2C communication
    i2c_master_stop(cmd);
    
    // Execute the I2C command
    esp_err_t ret = i2c_master_cmd_begin(i2c, cmd, 1000 / portTICK_PERIOD_MS);
    // Delete the I2C command link to free memory
    i2c_cmd_link_delete(cmd);

    return ret;
}

/*
Function:   Read I2C registers on the decibel sensor
NOTE:       Can be replaced with your custom I2C read code
Args:       i2c,        i2c port
            reg,        register address to write to
            data,       pointer to array to receive data into
            count,      number of bytes to read
Returns:    ESP_OK or ESP_FAIL
*/
esp_err_t i2cmaster_read_registers (i2c_port_t i2c, uint8_t reg, uint8_t *data, uint8_t count)
{
    // Use the same I2C port as the IO expander to avoid conflicts
    i2c_cmd_handle_t cmd = i2c_cmd_link_create();
    i2c_master_start(cmd);
    i2c_master_write_byte(cmd, (DBM_I2C_ADDR << 1) | I2C_MASTER_WRITE, true); // Write mode
    i2c_master_write_byte(cmd, reg, true);                         // Register address
    i2c_master_start(cmd);                                                     // Repeat start
    i2c_master_write_byte(cmd, (DBM_I2C_ADDR << 1) | I2C_MASTER_READ, true);  // Read mode
    
    // Read multiple bytes from the register
    if (count > 1)
        i2c_master_read(cmd, data, count - 1, I2C_MASTER_ACK); // Read n-1 bytes with ACK
    esp_err_t err = i2c_master_read_byte(cmd, &data[count - 1], I2C_MASTER_NACK); // Read last byte with NACK
    
    i2c_master_stop(cmd);
    err = i2c_master_cmd_begin(i2c, cmd, 1000 / portTICK_PERIOD_MS);
    i2c_cmd_link_delete(cmd);

    return err;
}

/*
Function:   Get module hardware version and ID
NOTE:       -
Args:       i2c,        i2c port
            dbm,        pointer to dbm_flex structure
Returns:    ESP_OK or ESP_FAIL
*/
esp_err_t dbmeter_get_version_id (i2c_port_t i2c, dbm_flex *dbm)
{
    if (i2cmaster_read_registers (i2c, DBM_REG_VERSION, &dbm->version, 1) != ESP_OK)
        return ESP_FAIL;
    if (i2cmaster_read_registers (i2c, DBM_REG_ID3, (uint8_t *)&dbm->id, 4) != ESP_OK)
        return ESP_FAIL;
    
    return ESP_OK;
}

/*
Function:   Update selected parameters by reading from the sensor
NOTE:       -
Args:       i2c,        i2c port
            dbm,        pointer to dbm_flex structure
            params,     ORed list of parameters to update
Returns:    ESP_OK or ESP_FAIL
*/
esp_err_t dbmeter_update_params (i2c_port_t i2c, dbm_flex *dbm, uint8_t params)
{
    if (params & DBM_PARAM_DB)
    {
        // Read time weighted decibel values
        uint8_t temp[6];
        if (i2cmaster_read_registers (i2c, DBM_REG_DBA_H, temp, 6) != ESP_OK)
            return ESP_FAIL;
        dbm->dBA = (float)temp[0] + (float)temp[1]/100.0f;
        dbm->dBC = (float)temp[2] + (float)temp[3]/100.0f;
        dbm->dBZ = (float)temp[4] + (float)temp[5]/100.0f;
    }

    if (params & DBM_PARAM_LEQfast)
    {
        // Read Leq fast values
        uint8_t temp[6];
        if (i2cmaster_read_registers (i2c, DBM_REG_LAEQF_H, temp, 6) != ESP_OK)
            return ESP_FAIL;
        dbm->LAeqfast = (float)temp[0] + (float)temp[1]/100.0f;
        dbm->LCeqfast = (float)temp[2] + (float)temp[3]/100.0f;
        dbm->LZeqfast = (float)temp[4] + (float)temp[5]/100.0f;
    }

    if (params & DBM_PARAM_LEQslow)
    {
        // Read Leq slow values
        uint8_t temp[6];
        if (i2cmaster_read_registers (i2c, DBM_REG_LAEQS_H, temp, 6) != ESP_OK)
            return ESP_FAIL;
        dbm->LAeqslow = (float)temp[0] + (float)temp[1]/100.0f;
        dbm->LCeqslow = (float)temp[2] + (float)temp[3]/100.0f;
        dbm->LZeqslow = (float)temp[4] + (float)temp[5]/100.0f;
    }

    if (params & DBM_PARAM_LPK)
    {
        // Read peak values
        uint8_t temp[6];
        if (i2cmaster_read_registers (i2c, DBM_REG_LAPK_H, temp, 6) != ESP_OK)
            return ESP_FAIL;
        dbm->LApk = (float)temp[0] + (float)temp[1]/100.0f;
        dbm->LCpk = (float)temp[2] + (float)temp[3]/100.0f;
        dbm->LZpk = (float)temp[4] + (float)temp[5]/100.0f;
    }

    if (params & DBM_PARAM_LMAXMIN)
    {
        // Read max/min values
        uint8_t temp[12];
        if (i2cmaster_read_registers (i2c, DBM_REG_LAMAX_H, temp, 12) != ESP_OK)
            return ESP_FAIL;
        dbm->LAmax = (float)temp[0] + (float)temp[1]/100.0f;
        dbm->LCmax = (float)temp[2] + (float)temp[3]/100.0f;
        dbm->LZmax = (float)temp[4] + (float)temp[5]/100.0f;
        dbm->LAmin = (float)temp[6] + (float)temp[7]/100.0f;
        dbm->LCmin = (float)temp[8] + (float)temp[9]/100.0f;
        dbm->LZmin = (float)temp[10] + (float)temp[11]/100.0f;
    }

    if (params & DBM_PARAM_TSTATS)
    {
        // Read over/under time values
        uint8_t temp[6];
        if (i2cmaster_read_registers (i2c, DBM_REG_T_OVER2, temp, 6) != ESP_OK)
            return ESP_FAIL;
        dbm->seconds_over = temp[0]<<16 | temp[1]<<8 | temp[2];
        dbm->seconds_under = temp[3]<<16 | temp[4]<<8 | temp[5];
    }

    return ESP_OK;
}

/*
Function:   Set Tavg or time weighting period
NOTE:       Minimum supported Tavg is 25 ms
Args:       i2c,        i2c port
            dbm,        pointer to dbm_flex structure, with desired Tavg value in it
Returns:    ESP_OK or ESP_FAIL
*/
esp_err_t dbmeter_set_tavg (i2c_port_t i2c, dbm_flex *dbm)
{
    uint8_t regdata = dbm->tavg >> 8;
    if (i2cmaster_write_registers (i2c, I2C_REG_TAVG_HIGH, &regdata, 1) != ESP_OK)
        return ESP_FAIL;
    regdata = dbm->tavg & 0xFF;
    if (i2cmaster_write_registers (i2c, I2C_REG_TAVG_LOW, &regdata, 1) != ESP_OK)
        return ESP_FAIL;

    return ESP_OK;
}

/*
Function:   Enters and exits sleep mode
NOTE:       Module settings are reset on wakeup
Args:       i2c,                i2c port
            activate_sleep,     1 for sleep, 0 for wakeup
Returns:    ESP_OK or ESP_FAIL
*/
esp_err_t dbmeter_set_sleep (i2c_port_t i2c, uint8_t activate_sleep)
{
    uint8_t regdata;

    if (activate_sleep)
    {
        // Put module into sleep mode
        regdata = 0x01;
        if (i2cmaster_write_registers (i2c, I2C_REG_CONTROL, &regdata, 1) != ESP_OK)
            return ESP_FAIL;
    }
    else
    {
        // Wake up module by triggering reset
        regdata = 0x20;
        if (i2cmaster_write_registers (i2c, I2C_REG_RESET, &regdata, 1) != ESP_OK)
            return ESP_FAIL;
    }

    return ESP_OK;
}

/*
Function:   Clears selected parameters from the module
NOTE:       Call this after power-up to clear max/min caused by power-up
            dbm_flex structure must be updated after clearing values in module
Args:       i2c,        i2c port
            params,     ORed DBM_CLR_x flags to clear corresponding parameters
Returns:    ESP_OK or ESP_FAIL
*/
esp_err_t dbmeter_clear_data (i2c_port_t i2c, uint8_t params)
{
    if (i2cmaster_write_registers (i2c, I2C_REG_RESET, &params, 1) != ESP_OK)
        return ESP_FAIL;
    return ESP_OK;
}

/*
Function:   Reads 60 LAeq,1s values recorded over the last 1 minute
NOTE:       -
Args:       i2c,        i2c port
            history,    buffer to populate with history values, must be 60 units long
Returns:    ESP_OK or ESP_FAIL
*/
esp_err_t dbmeter_read_history_sec (i2c_port_t i2c, uint8_t *history)
{
    if (i2cmaster_read_registers (i2c, DBM_REG_HISTORY_SEC, history, 60) != ESP_OK)
        return ESP_FAIL;
    else
        return ESP_OK;
}

/*
Function:   Reads 60 LAeq,1min values recorded over the last 1 hour
NOTE:       -
Args:       i2c,        i2c port
            history,    buffer to populate with history values, must be 60 units long
Returns:    ESP_OK or ESP_FAIL
*/
esp_err_t dbmeter_read_history_min (i2c_port_t i2c, uint8_t *history)
{
    if (i2cmaster_read_registers (i2c, DBM_REG_HISTORY_MIN, history, 60) != ESP_OK)
        return ESP_FAIL;
    else
        return ESP_OK;
}

/*
Function:   Reads 60 LAeq,1hr values recorded over the last 12 hours
NOTE:       -
Args:       i2c,        i2c port
            history,    buffer to populate with history values, must be 60 units long
Returns:    ESP_OK or ESP_FAIL
*/
esp_err_t dbmeter_read_history_hrs (i2c_port_t i2c, uint8_t *history)
{
    if (i2cmaster_read_registers (i2c, DBM_REG_HISTORY_HRS, history, 12) != ESP_OK)
        return ESP_FAIL;
    else
        return ESP_OK;
}
